// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file audio.cc
/// @brief Class for libao wrapping.

#include "../include/audio.h"
#include "../include/mainmanager.h"
#include "../include/trace.h"
#include "../include/misc.h"

std::auto_ptr<Audio> Audio::audio_;

Audio::Audio(MainManager& mainManager)
: bufferCount_(0),
doubleInverseRandMax_( 2.0 / RAND_MAX ),
ditherRand_(0.0),
ditherTemp_(0.0),
ditherXe_(0.0),
ditherResult_(0.0),
mainManager_(mainManager),    
device_(NULL),
option_(NULL),
driverId_(-1),
playLoopThread_(NULL),
playLoopStart_(false),
playLoopStop_(false),
playLoopPause_(false),
skipBufferFlush_(false),
singleSampleSize_(0),
bufferSize_(0),
littleEndian_(true),
tempScaledSample_(0)
{
    memset(buffer_, 0, 142848);
     
    ao_initialize();

    SetDriverFromPreference();
}

Audio::~Audio()
{
    ao_free_options(option_);
    ao_close(device_);
    ao_shutdown();

}

/// @brief Get driver id and option from preference, prepare to that.
///
/// If audio driver id is "default", try to set "esd" first.
void Audio::SetDriverFromPreference()
{
    Tracer::Trace trace("Audio::SetDriverFromPreference");

    if(CheckOpenDevice()) 
	Close();

    int defaultDriver = ao_default_driver_id();
    std::string audioDriver = mainManager_.GetPreference().AudioDriver();
    // Don't worry, if pass NULL to ao_free_options, do nothing!
    ao_free_options(option_);
    option_ = NULL;

    if(audioDriver == "default") {
	if(TestDriver("esd")) {
	    trace.Debug("Detect esd on default driver");
	    
	    driverId_ = ao_driver_id("esd");

	    return;
	}
	    
	trace.Debug("Driver is default");

	driverId_ = defaultDriver;

    } else	
	driverId_ = ao_driver_id(audioDriver.c_str());

    if(driverId_ == -1) {
	trace.Debug("Got error, select default driver!");

	driverId_ = defaultDriver;	

    } else {
	trace.Debug("Selected not default driver : ");
	trace.Debug(audioDriver.c_str());

	if(audioDriver == "alsa") {
	    ao_append_option(&option_, "card", mainManager_.GetPreference().AudioAlsa5Card().c_str());
	    ao_append_option(&option_, "dev", mainManager_.GetPreference().AudioAlsa5Dev().c_str());	

	} else if(audioDriver == "alsa09") {
	    ao_append_option(&option_, "dev", mainManager_.GetPreference().AudioAlsa9Dev().c_str());

	} else if(audioDriver == "oss") {
	    ao_append_option(&option_, "dsp", mainManager_.GetPreference().AudioOssDsp().c_str());
	    
	} else if(audioDriver == "sun") {
	    ao_append_option(&option_, "dev", mainManager_.GetPreference().AudioSunDev().c_str());

	} else if(audioDriver == "aixs") {
	    ao_append_option(&option_, "dev", mainManager_.GetPreference().AudioAixsDev().c_str());

	}
    }
}

/// @brief Test to open device from driver name.
///
/// Even driver name is "default", does not try to "esd" first.
/// @param driverName Driver name to test.
/// @return Sucess of test.
bool Audio::TestDriver(const char* const driverName)
{
    Tracer::Trace trace("Audio::TestDriver without option");

    trace.Debug("Selected driver : ");
    trace.Debug(driverName);
    
    int driverId = ao_default_driver_id();

    if(strncmp(driverName, "default", 3) != 0) {
	driverId = ao_driver_id(driverName);
    }

    if(driverId == -1) {
	trace.Debug("Can't find such driver id");
	
	return false;
    }

    ao_sample_format format;
    format.bits = 16;
    format.channels = 2;
    format.rate = 44100;
    format.byte_format = AO_FMT_NATIVE;

    DeviceManager device(driverId, &format, NULL);
    
    return device.Valid();

}

/// @brief Test to open device from driver name.
///
/// If audio driver id is "default", try to set "esd" first.
/// @param driverName Driver name to test.
/// @param option Driver option.
/// @return Sucess of test.
bool Audio::TestDriver(const char* const driverName, const std::map<std::string, std::string>& option)
{
    Tracer::Trace trace("Audio::TestDriver");

    trace.Debug("Selected driver : ");
    trace.Debug(driverName);
    
    int driverId = ao_default_driver_id();

    if(strncmp(driverName, "default", 3) == 0) {
	if(TestDriver("esd"))
	    return true;
    } else
	driverId = ao_driver_id(driverName);

    if(driverId == -1) {
	trace.Debug("Can't find such driver id");
	
	return false;
    }

    OptionManager aoOption;
    if(!option.empty()) {

	for_each(option.begin(), option.end(), std::bind2nd(AoOptionAppend(), &aoOption));

    }

    ao_sample_format format;
    format.bits = 16;
    format.channels = 2;
    format.rate = 44100;
    format.byte_format = AO_FMT_NATIVE;

    DeviceManager device(driverId, &format, aoOption.Get());

    return device.Valid();

}

/// @brief Open device and start play buffer thread.
///
/// If bits over 16 or channels more than 2, return false.
/// @param bits Bits per sample.
/// @param channels Number of channels.
/// @param sampleRate Sample rate per sample.
/// @param byteFormat Byte order. AO_FMT_LITTLE (little endian), AO_FMT_BIG (big endian), AO_FMT_NATIVE (machine's native endian). 
/// @return Success of open.
bool Audio::OpenDevice(const unsigned int& bits, const unsigned int& channels, const unsigned int& sampleRate, const unsigned int& byteFormat)
{
    if(bits != 16 || channels > 2)
	// Can't control if that isn't 16 bits or more than 2 channels Now.
	return false;

    Reset();
    Effect::Reset();	
	
    format_.bits = bits;
    format_.channels = channels;
    format_.rate = sampleRate;
    format_.byte_format = byteFormat;

    singleSampleSize_ = bits / 8 * channels;
    switch(byteFormat) {
	case AO_FMT_NATIVE:
	    if(IsBigEndian())
		littleEndian_ = false;
	    else
		littleEndian_ = true;
			 
	    break;

	case AO_FMT_BIG:
	    littleEndian_ = false;

	    break;
	    
	case AO_FMT_LITTLE:
	default:
	    littleEndian_ = true;
    }
    
    device_ = ao_open_live(driverId_, &format_, option_);

    if(device_) {
	
	playLoopThread_ = Glib::Thread::create( 
	    sigc::mem_fun(*this, &Audio::PlayBufferLoop), true);

	playLoopStart_ = true;

	return true;
    
    } else
	return false;	
}

// Can you gurantee call Play after OpenDevice ?
/// @brief Process effector and dithering sample buffer, push to play buffer.
///
/// samples[channels][signal]
///
///
/// ex) samples[0][0] -> left channel at 0, samples[0][3] -> left at channel 3.
///
/// samples[1][0] -> right channel at 0, samples[1][5] -> right channel at 5.
/// @param samples Sample buffer.
/// @param bytes Sample buffer size.
void Audio::Play(float** const samples, const unsigned int& bytes)
{
    bufferSize_ = singleSampleSize_ * bytes;

    char tmpBuf[bufferSize_];
    
    for(unsigned int i = 0; i < bytes; ++i)
    {
	// samples[channels][signal]
	//
	// ex.  samples[0][0] -> left at 0
	//	    samples[1][0] -> right at 0
	//
	// becareful index boundry when it is not regular signal. (mono, 
	// mutiple channels... )

	#warning "Change this index if source has multiple channels"
	stereoSignal_.left = static_cast<double>(samples[0][i]);
	if(format_.channels == 1)
	    stereoSignal_.right = 0.0;
	else    
	    stereoSignal_.right = static_cast<double>(samples[1][i]);
	    
	if(mainManager_.GetPreference().Effect())    
	    Effect::Process(mainManager_.GetPreference(), format_.rate, stereoSignal_);

	ScaleTo16Bit(stereoSignal_.left, dLeft_);
	ShortIntToChar(tempScaledSample_, &tmpBuf[singleSampleSize_*i]);

	if(format_.channels == 2) {
	    ScaleTo16Bit(stereoSignal_.right, dRight_);
	    ShortIntToChar(tempScaledSample_, &tmpBuf[singleSampleSize_*i+2]);
	}
    }

    PlayBufferPush(tmpBuf);

    while(playLoopPause_)
	usecsleep(0, 100);
    
}

/// @brief Process effector and dithering sample buffer, push to play buffer.
///
/// In sample is stereo, even index is left channel, and odd index is right channel.
///
///
/// ex) samples[0] -> left channel at 0, samples[2] -> left channel at 1
///
/// samples[1] -> right channel at 0, samples[3] -> right channel at 1
/// @param samples Sample buffer.
/// @param bytes Sample buffer size.
void Audio::Play(const std::vector<double>& samples, const unsigned int& bytes)
{
    bufferSize_ = singleSampleSize_ * bytes;

    char tmpBuf[bufferSize_];
    
    for(unsigned int i = 0; i < bytes; ++i)
    {
	stereoSignal_.left = samples[i*2];

	// i initialized right signal 0.0 when it is not stereo.
	// But who knows? anyway i'll assign 0.0 again....
	if(format_.channels == 1)
	    stereoSignal_.right = 0.0;
	else    
	    stereoSignal_.right = samples[i*2 + 1]; 

	if(mainManager_.GetPreference().Effect())
	    Effect::Process(mainManager_.GetPreference(), format_.rate, stereoSignal_);

	ScaleTo16Bit(stereoSignal_.left, dLeft_);
	ShortIntToChar(tempScaledSample_, &tmpBuf[singleSampleSize_*i]);
	
	if(format_.channels == 2) {
	    ScaleTo16Bit(stereoSignal_.right, dRight_);
	    ShortIntToChar(tempScaledSample_, &tmpBuf[singleSampleSize_*i+2]);
	}
    }

    PlayBufferPush(tmpBuf);

    while(playLoopPause_)
	usecsleep(0, 100);
    
}

/// @brief Play buffered samples.
///
/// Pop samples from play buffer and copy to local buffer. After that
/// play local buffer.
void Audio::PlayBufferLoop()
{
    char tmpBuffer[32768];

    for(;;) {

	{
	    Glib::Mutex::Lock lock(playLoopMutex_);

	    if(bufferCount_ < 32768) {
		while(bufferCount_ < 131072 && !playLoopStop_) 
		    playLoopCondPush_.wait(playLoopMutex_);
	    }
	 
	    if(playLoopStop_)
		break;
	    
	    memcpy(tmpBuffer, buffer_, 32768);
	    
	    bufferCount_ -= 32768;
	    
	    memcpy(buffer_, buffer_ + 32768, bufferCount_);	
	    
	    playLoopCondPop_.signal();
	}
	
	if(!skipBufferFlush_)	 
	    ao_play(device_, tmpBuffer, 32768);

	while(playLoopPause_)
	    usecsleep(0, 100);

    }
}

/// @brief Dithering to 16 bit.
///
/// Dithering result is member variable tempScaledSample_.
/// @param sample Sample to dither.
/// @param dither Dither variable.
void Audio::ScaleTo16Bit(const double& sample, Dither& dither)
{
    ditherRand_ = static_cast<double>(rand());
    ditherRand_ *= doubleInverseRandMax_;
    ditherRand_ -= 1.0;
    
    ditherXe_ = sample;
    ditherXe_ *= 32767.5;
    
    ditherTemp_ = 0.4191;
    ditherTemp_ *= dither.dBuffer_[Mod((dither.dPos_+8), 9)];
    ditherXe_ -= ditherTemp_;
    
    ditherTemp_ = 1.632;
    ditherTemp_ *= dither.dBuffer_[Mod((dither.dPos_+7), 9)];
    ditherXe_ += ditherTemp_;

    ditherTemp_ = 3.263;
    ditherTemp_ *= dither.dBuffer_[Mod((dither.dPos_+6), 9)];
    ditherXe_ -= ditherTemp_;

    ditherTemp_ = 5.032;
    ditherTemp_ *= dither.dBuffer_[Mod((dither.dPos_+5), 9)];
    ditherXe_ += ditherTemp_;

    ditherTemp_ = 6.639;
    ditherTemp_ *= dither.dBuffer_[Mod((dither.dPos_+4), 9)];
    ditherXe_ -= ditherTemp_;

    ditherTemp_ = 7.184;
    ditherTemp_ *= dither.dBuffer_[Mod((dither.dPos_+3), 9)];
    ditherXe_ += ditherTemp_;

    ditherTemp_ = 6.214;
    ditherTemp_ *= dither.dBuffer_[Mod((dither.dPos_+2), 9)];
    ditherXe_ -= ditherTemp_;

    ditherTemp_ = 4.684;
    ditherTemp_ *= dither.dBuffer_[Mod((dither.dPos_+1), 9)];
    ditherXe_ += ditherTemp_;

    ditherTemp_ = 2.847;
    ditherTemp_ *= dither.dBuffer_[(dither.dPos_)];
    ditherXe_ -= ditherTemp_;

    ditherResult_ = ditherXe_;
    ditherResult_ += ditherRand_;
    ditherResult_ -= dither.dTState_;
    
    dither.dPos_ = Mod((dither.dPos_+8), 9);
    dither.dBuffer_[dither.dPos_] = ditherXe_ - rint(ditherResult_);
    dither.dTState_ = ditherRand_;
    
    // In linux man page, rint conforms to BSD 4.3 not C99,
    // but there exist rint in C99. What is truth?
    // Anyway we can use rint in linux and BSD both.
    ditherResult_ = rint(ditherResult_);
    
    if(ditherResult_ > 32767.0)
	tempScaledSample_ = 32767;
    else if(ditherResult_ < -32768.0)
	tempScaledSample_ = -32768;
    else
	tempScaledSample_ = static_cast<short int>(ditherResult_);
    
}

/// @brief Reset member variables.
void Audio::Reset()
{
    ditherRand_ = 0.0;
    ditherTemp_ = 0.0;
    ditherXe_ = 0.0;
    ditherResult_ = 0.0;
    
    tempScaledSample_ = 0;

    stereoSignal_.left = 0.0;
    stereoSignal_.right = 0.0;

    dLeft_.dTState_ = 0.0;
    dLeft_.dPos_ = 8;

    dRight_.dTState_ = 0.0;
    dRight_.dPos_ = 8;

    for(unsigned int i = 0; i < 9; ++i)
    {
	dLeft_.dBuffer_[i] = 0.0;
	dRight_.dBuffer_[i] = 0.0;
    }

}

