// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file imagebutton.cc
/// @brief Class for rollover image button source file.

#include "../include/imagebutton.h"

/// @brief Constructor of ImageButton.
///
/// Assume all button images have same size, 24 pixel height.
/// @param normal Button normal image file path.
/// @param over Button over image file path.
/// @param press Button press image file path.
/// @todo Detect button image size.
ImageButton::ImageButton(const std::string& normal, const std::string& over, const std::string& press)
    : normalImage_(normal),
    overImage_(over),
    pressImage_(press)
{
    add(mainBox_);
  
    normalImage_.show();
    prevImage_ = &normalImage_;

    overImage_.set_child_visible(false);
    overImage_.hide();

    pressImage_.set_child_visible(false);
    pressImage_.hide();

    // well.. not good magic constant...
    mainBox_.set_size_request(-1, 24); 
    mainBox_.pack_start(normalImage_, Gtk::PACK_SHRINK);
    mainBox_.pack_start(overImage_, Gtk::PACK_SHRINK);
    mainBox_.pack_start(pressImage_, Gtk::PACK_SHRINK);

    set_events(Gdk::ENTER_NOTIFY_MASK | Gdk::LEAVE_NOTIFY_MASK | Gdk::BUTTON_PRESS_MASK | Gdk::BUTTON_RELEASE_MASK);    

    signal_enter_notify_event().connect(
	sigc::mem_fun(*this, &ImageButton::OnEnterNotify) ); 
    signal_leave_notify_event().connect(
	sigc::mem_fun(*this, &ImageButton::OnLeaveNotify) );
    signal_button_press_event().connect(
	sigc::mem_fun(*this, &ImageButton::OnBtnPress) );
    signal_button_release_event().connect(
	sigc::mem_fun(*this, &ImageButton::OnBtnRelease) );	
}

ImageButton::~ImageButton()
{}

/// @brief You need to connect this signal to detect button is pressed.
sigc::signal<void> ImageButton::SignalPress()
{
    return signalPress_;
}

/// @brief Callback function when button is pressed.
///
/// If pressed button is 1 button, hide previous image and display pressImage.
/// And transmit signal to connected to signalPress. 
bool ImageButton::OnBtnPress(GdkEventButton* ev)
{
    if( (ev->type == GDK_BUTTON_PRESS) && (ev->button == 1) ) {
	prevImage_->set_child_visible(false);
	prevImage_->hide();
	prevImage_ = &pressImage_;
	pressImage_.set_child_visible(true);
	pressImage_.show();

	signalPress_.emit();
    }

    return false;
}

/// @brief Callback function when button is released.
bool ImageButton::OnBtnRelease(GdkEventButton* ev)
{
    prevImage_->set_child_visible(false);
    prevImage_->hide();
    prevImage_ = &overImage_;
    overImage_.set_child_visible(true);
    overImage_.show();

    return false;
}

/// @brief Callback function when button is entered notify.
bool ImageButton::OnEnterNotify(GdkEventCrossing* ev)
{
    prevImage_->set_child_visible(false);
    prevImage_->hide();
    prevImage_ = &overImage_;
    overImage_.set_child_visible(true);
    overImage_.show();

    return false;
}

/// @brief Callback function when button is leaved notify.
bool ImageButton::OnLeaveNotify(GdkEventCrossing* ev)
{
    prevImage_->set_child_visible(false);
    prevImage_->hide();
    prevImage_ = &normalImage_;
    normalImage_.set_child_visible(true);
    normalImage_.show();

    return false;
}

